from pykechain.exceptions import _DeprecationMixin


class SelectListProperty2(_DeprecationMixin):
    """A virtual object representing a KE-chain single-select list property."""

    pass


class MultiSelectListProperty2(_DeprecationMixin):
    """A virtual object representing a KE-chain multi-select list property."""

    pass
