from pykechain.exceptions import _DeprecationMixin


class ActivityReferencesProperty(_DeprecationMixin):
    """A virtual object representing a KE-chain Activity References property.

    .. versionadded:: 3.7
    """

    pass
