from pykechain.exceptions import _DeprecationMixin


class MultiReferenceProperty2(_DeprecationMixin):
    """A virtual object representing a KE-chain multi-references property.

    .. versionadded:: 1.14
    """

    pass
